@@ -1,5 +1,15 @@
 /* -*- indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */
 
+/* Copyright (c) 2004 - 2006 Derek Foreman, Ben Jansens
+   Copyright (c) 2006 - 2012 Thomas Schmitt <scdbackup@gmx.net>
+   Provided under GPL version 2 or later.
+*/
+
+#ifdef HAVE_CONFIG_H
+#include "../config.h"
+#endif
+
+
 #include <unistd.h>
 
 /* ts A61007 */
@@ -9,6 +19,7 @@
 #include <signal.h>
 #include <string.h>
 #include <stdlib.h>
+#include <pthread.h>
 
 /* ts A70928 : init.h is for others, not for init .c
 #include "init.h"
@@ -32,14 +43,16 @@ struct libdax_msgs *libdax_messenger= NULL;
 
 int burn_running = 0;
 
-/* ts A60813 : Linux: wether to use O_EXCL on open() of device files */
+/* ts A60813 : GNU/Linux: whether to use O_EXCL on open() of device files
+   ts B00212 : FreeBSD:   whether to use flock(LOCK_EX) after open()
+*/
 int burn_sg_open_o_excl = 1;
 
-/* ts A70403 : Linux: wether to use fcntl(,F_SETLK,)
+/* ts A70403 : GNU/Linux: wether to use fcntl(,F_SETLK,)
                       after open() of device files  */
 int burn_sg_fcntl_f_setlk = 1;
 
-/* ts A70314 : Linux: what device family to use :
+/* ts A70314 : GNU/Linux: what device family to use :
     0= default family
     1= sr
     2= scd
@@ -59,6 +72,12 @@ int burn_sg_open_o_nonblock = 1;
 	    to unconditional abort of the process  */
 int burn_sg_open_abort_busy = 0;
 
+
+/* The message returned from sg_id_string() and/or sg_initialize()
+*/
+static char sg_initialize_msg[1024] = {""};
+
+
 /* ts A61002 */
 
 #include "cleanup.h"
@@ -66,15 +85,33 @@ int burn_sg_open_abort_busy = 0;
 /* Parameters for builtin abort handler */
 static char abort_message_prefix[81] = {"libburn : "};
 static pid_t abort_control_pid= 0;
+static pthread_t abort_control_thread;
 volatile int burn_global_abort_level= 0;
 int burn_global_abort_signum= 0;
 void *burn_global_signal_handle = NULL;
 burn_abort_handler_t burn_global_signal_handler = NULL;
+int burn_builtin_signal_action = 0;            /* burn_set_signal_handling() */
+volatile int burn_builtin_triggered_action = 0;       /*  burn_is_aborting() */
 
 
 /* ts A70223 : wether implemented untested profiles are supported */
 int burn_support_untested_profiles = 0;
 
+/* ts A91111 :
+   whether to log SCSI commands (to be implemented in sg-*.c)
+   bit0= log in /tmp/libburn_sg_command_log
+   bit1= log to stderr
+   bit2= flush every line
+*/
+int burn_sg_log_scsi = 0;
+
+
+/* ts B10312 :
+   Whether to map random-access readonly files to drive role 4.
+   Else it is role 2 overwriteable drive
+*/
+int burn_drive_role_4_allowed = 0;
+
 
 /* ts A60925 : ticket 74 */
 /** Create the messenger object for libburn. */
@@ -103,6 +140,14 @@ int burn_initialize(void)
 	ret = burn_msgs_initialize();
 	if (ret <= 0)
 		return 0;
+	ret = sg_initialize(sg_initialize_msg, 0);
+	if (ret <= 0) {
+                libdax_msgs_submit(libdax_messenger, -1,
+                        0x00020175,
+                        LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+                        sg_initialize_msg, 0, 0);
+		return 0;
+	}
 	burn_running = 1;
 	return 1;
 }
@@ -130,10 +175,24 @@ void burn_finish(void)
 	/* ts A60924 : ticket 74 */
 	libdax_msgs_destroy(&libdax_messenger,0);
 
+	sg_shutdown(0);
+
+	burn_drive_clear_whitelist();
+
 	burn_running = 0;
 }
 
 
+/* ts A91226 */
+/** API function. See libburn.h */
+char *burn_scsi_transport_id(int flag)
+{
+	if (!burn_running)
+		sg_id_string(sg_initialize_msg, 0);
+	return sg_initialize_msg;
+}
+
+
 /* ts A60813 */
 /** API function. See libburn.h */
 void burn_preset_device_open(int exclusive, int blocking, int abort_on_busy)
@@ -288,6 +347,37 @@ int burn_sev_to_text(int severity_number, char **severity_name, int flag)
 	return ret;
 }
 
+/* ts B00224 */
+char *burn_util_thread_id(pid_t pid, pthread_t tid, char text[80])
+{
+	int i, l;
+
+	sprintf(text, "[%lu,", (unsigned long int) getpid());
+	l= strlen(text);
+	for(i= 0; i < ((int) sizeof(pthread_t)) && 2 * i < 80 - l - 3; i++)
+		sprintf(text + l + 2 * i,
+			 "%2.2X", ((unsigned char *) &tid)[i]);
+
+	sprintf(text + l + 2 * i, "]");
+	return text;
+}
+
+
+/* ts B20122 */
+/* @param value 0=return rather than exit(value)
+*/
+int burn_abort_exit(int value)
+{
+	burn_abort(4440, burn_abort_pacifier, abort_message_prefix);
+	fprintf(stderr,
+	"\n%sABORT : Program done. Even if you do not see a shell prompt.\n\n",
+		abort_message_prefix);
+	if (value)
+		exit(value);
+	burn_global_abort_level = -2;
+	return(1);
+}
+
 
 int burn_builtin_abort_handler(void *handle, int signum, int flag)
 {
@@ -296,15 +386,53 @@ int burn_builtin_abort_handler(void *handle, int signum, int flag)
 /*
 #define Libburn_signal_handler_verbouS 1
 */
+
 	int ret;
 	struct burn_drive *d;
 
 #ifdef Libburn_signal_handler_verbouS
-	fprintf(stderr,
-		"libburn_ABORT: pid = %d , abort_control_pid = %d , sig= %d\n",
-		getpid(), abort_control_pid, signum);
+	char text[80];
+
+	fprintf(stderr, "libburn_ABORT: in = %s\n",
+		burn_util_thread_id(getpid(), pthread_self(), text));
+	fprintf(stderr, "libburn_ABORT: ctrl = %s\n",
+		burn_util_thread_id(abort_control_pid, abort_control_thread,
+					text));
+	if (burn_global_signal_handler == burn_builtin_abort_handler)
+		fprintf(stderr, "libburn_ABORT: signal action = %d\n",
+				burn_builtin_signal_action);
+
+	/* >>> find writing drives and report their tid
+	fprintf(stderr, "libburn_ABORT: wrt = %s\n",
+		burn_util_thread_id(0, burn_write_thread_id, text));
+	fprintf(stderr, "libburn_ABORT: sig= %d\n", signum);
+	*/
 #endif
 
+	burn_builtin_triggered_action = burn_builtin_signal_action;
+	burn_global_abort_level = -1;
+
+	if (burn_builtin_signal_action > 1) {
+		Cleanup_set_handlers(NULL, NULL, 2);
+		if (burn_builtin_signal_action == 4)
+			return -2;
+		fprintf(stderr,"%sABORT : Trying to shut down busy drives\n",
+			abort_message_prefix);
+		fprintf(stderr,
+		 "%sABORT : Wait the normal burning time before any kill -9\n",
+		 	abort_message_prefix);
+		burn_abort_5(0, burn_abort_pacifier, abort_message_prefix,
+				0, 1);
+		libdax_msgs_submit(libdax_messenger, -1, 0x00020177,
+			LIBDAX_MSGS_SEV_ABORT, LIBDAX_MSGS_PRIO_HIGH,
+			"Urged drive worker threads to do emergency halt",
+			0, 0);
+		return -2;
+	}
+
+
+	/* ---- old deprecated stuck-in-abort-handler loop ---- */
+
 	/* ts A70928:
 	Must be quick. Allowed to coincide with other thread and to share
 	the increment with that one. It must not decrease, though, and
@@ -317,7 +445,8 @@ int burn_builtin_abort_handler(void *handle, int signum, int flag)
 
 #ifdef Libburn_new_thread_signal_handleR
 
-		ret = burn_drive_find_by_thread_pid(&d, getpid());
+		ret = burn_drive_find_by_thread_pid(&d, getpid(),
+							pthread_self());
 		if (ret > 0 && d->busy == BURN_DRIVE_WRITING) {
 					/* This is an active writer thread */
 
@@ -357,6 +486,7 @@ int burn_builtin_abort_handler(void *handle, int signum, int flag)
 	}
 	burn_global_abort_level = -1;
 	Cleanup_set_handlers(NULL, NULL, 2);
+
 	fprintf(stderr,"%sABORT : Trying to shut down drive and library\n",
 		abort_message_prefix);
 	fprintf(stderr,
@@ -364,37 +494,110 @@ int burn_builtin_abort_handler(void *handle, int signum, int flag)
 		abort_message_prefix);
 	close(0); /* somehow stdin as input blocks abort until EOF */
 
-	burn_abort(4440, burn_abort_pacifier, abort_message_prefix);
-
-	fprintf(stderr,
-	"\n%sABORT : Program done. Even if you do not see a shell prompt.\n\n",
-		abort_message_prefix);
-	burn_global_abort_level = -2;
-	return(1);
+	burn_abort_exit(0);
+	return (1);
 }
 
+
+/* ts A61002 : API */
 void burn_set_signal_handling(void *handle, burn_abort_handler_t handler,
-			     int mode)
+				int mode)
 {
-	if(handler == NULL && mode == 0) {
+
+/*
+	fprintf(stderr, "libburn_experimental: burn_set_signal_handling, handler==%lx  mode=%d\n", (unsigned long) handler, mode);
+*/
+
+	if(handler == NULL) {
 		handler = burn_builtin_abort_handler;
 /*
-		fprintf(stderr, "libburn_experimental: activated burn_builtin_abort_handler() with handle '%s'\n",(handle==NULL ? "libburn : " : (char *) handle));
+		if ((mode & ~4) == 0)
+			fprintf(stderr, "libburn_experimental: activated burn_builtin_abort_handler() with handle '%s'\n",(handle==NULL ? "libburn : " : (char *) handle));
 */
 
 	}
 	strcpy(abort_message_prefix, "libburn : ");
-	if(handle != NULL)
+	abort_message_prefix[0] = 0;
+	if(handle != NULL && handler == burn_builtin_abort_handler)
 		strncpy(abort_message_prefix, (char *) handle,
 			sizeof(abort_message_prefix)-1);
 	abort_message_prefix[sizeof(abort_message_prefix)-1] = 0;
 	abort_control_pid = getpid();
-	Cleanup_set_handlers(handle, (Cleanup_app_handler_T) handler, mode|4);
+	abort_control_thread = pthread_self();
+	burn_builtin_signal_action = (mode >> 4) & 15;
+	if((mode & 11) != 0)
+		burn_builtin_signal_action = 0;
+	if(burn_builtin_signal_action > 1)
+		burn_builtin_triggered_action = 0;
+	if(burn_builtin_signal_action == 0)
+		burn_builtin_signal_action = 1;
+	Cleanup_set_handlers(handle, (Cleanup_app_handler_T) handler,
+				 (mode & 15) | 4);
 	burn_global_signal_handle = handle;
 	burn_global_signal_handler = handler;
 }
 
 
+/* ts B00304 : API */
+int burn_is_aborting(int flag)
+{
+	return burn_builtin_triggered_action;
+}
+
+
+/* ts B00225 */
+/* @return 0= no abort action 2 pending , 1= not control thread
+*/
+int burn_init_catch_on_abort(int flag)
+{
+	if (burn_builtin_triggered_action != 2)
+		return 0;
+	if (abort_control_pid != getpid() ||
+		abort_control_thread != pthread_self())
+		return 1;
+	burn_abort(4440, burn_abort_pacifier, abort_message_prefix);
+	fprintf(stderr,
+	"\n%sABORT : Program done. Even if you do not see a shell prompt.\n\n",
+		abort_message_prefix);
+	exit(1);
+}
+
+
+/* B20122 */
+/* Temporarily disable builtin actions 0,1,2 to avoid that burn_abort()
+   waits for its own thread to end grabbing.
+*/
+int burn_grab_prepare_sig_action(int *signal_action_mem, int flag)
+{
+	*signal_action_mem = -1;
+	if (burn_global_signal_handler == burn_builtin_abort_handler &&
+	    burn_builtin_signal_action >= 0 &&
+	    burn_builtin_signal_action <= 2) {
+		*signal_action_mem = burn_builtin_signal_action;
+		burn_builtin_signal_action = 3;
+	}
+	return 1;
+}
+
+
+/* B20122 */
+/* Re-enable builtin actions 0,1,2 and perform delayed signal reactions
+*/
+int burn_grab_restore_sig_action(int signal_action_mem, int flag)
+{
+	if (signal_action_mem >= 0)
+		burn_builtin_signal_action = signal_action_mem;
+	if (burn_is_aborting(0) && signal_action_mem >= 0) {
+		if (signal_action_mem == 0 || signal_action_mem == 1) {
+			burn_abort_exit(1); /* Never comes back */
+		} else if (signal_action_mem == 2) {
+			burn_builtin_triggered_action = signal_action_mem;
+		}
+	}
+	return 1;
+}
+
+
 /* ts A70223 : API */
 void burn_allow_untested_profiles(int yes)
 {
@@ -414,3 +617,31 @@ int burn_set_messenger(void *messenger)
 	return 1;
 }
 
+
+/* ts A91111 API */
+void burn_set_scsi_logging(int flag)
+{
+	burn_sg_log_scsi = flag & 7;
+}
+
+
+/* ts B10312 API */
+void burn_allow_drive_role_4(int allowed)
+{
+	burn_drive_role_4_allowed = (allowed & 0xf);
+}
+
+
+/* ts B10606 */
+void *burn_alloc_mem(size_t size, size_t count, int flag)
+{
+	void *pt;
+
+	pt = calloc(count, size);
+	if(pt == NULL)
+		libdax_msgs_submit(libdax_messenger, -1, 0x00000003,
+				LIBDAX_MSGS_SEV_FATAL, LIBDAX_MSGS_PRIO_HIGH,
+				"Out of virtual memory", 0, 0);
+	return pt;
+}
+
